
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/rumd_base.h"
#include "rumd/rumd_utils.h"
#include "rumd/AxilrodTellerPotential.h"
#include "rumd/NeighborList.h"

template<class S>
__global__ void calcf_AxTe(unsigned int num_part, unsigned int nvp, __restrict__ float4* r, __restrict__ float4* f, float4* w, S* simBox, float* simBoxPointer, float v_AT, unsigned *num_nbrs, unsigned *nbl);


AxilrodTellerPotential::AxilrodTellerPotential(float v_AT, float Rcut) :
  allocated_size_pe(0) {
  this->v_AT = v_AT;
  this->Rcut = Rcut;
  SetID_String("potAxTe");
}


void AxilrodTellerPotential::AllocatePE_Array(unsigned int nvp) {
  if(allocated_size_pe != 0) {
    cudaFreeHost(h_f_w_loc);
    cudaFree(d_f_loc);
    cudaFree(d_w_loc);
  }
  
  if( cudaMallocHost( (void**) &h_f_w_loc, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("AxilrodTellerPotential", __func__, "Malloc failed on h_f_w_loc") );
  if( cudaMalloc( (void**) &d_f_loc, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("AxilrodTellerPotential", __func__, "Malloc failed on d_f_loc") );
  if( cudaMalloc( (void**) &d_w_loc, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("AxilrodTellerPotential", __func__, "Malloc failed on d_w_loc") );

  memset( h_f_w_loc, 0, nvp * sizeof(float4) );
  cudaMemset( d_f_loc, 0, nvp * sizeof(float4) );
  cudaMemset( d_w_loc, 0, nvp * sizeof(float4) );
  
  allocated_size_pe = nvp;
}


void AxilrodTellerPotential::Initialize() {

 
  neighborList.Initialize(sample);
  
  neighborList.SetEqualCutoffs(1, Rcut);

  testRSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  testLESB = dynamic_cast<LeesEdwardsSimulationBox*>(simBox);
  if (!testRSB && !testLESB) throw RUMD_Error("AxilrodTellerPotential", __func__, "simBox must be either RectangularSimulationBox or LeesEdwardsSimulationBox");

  
}

void AxilrodTellerPotential::CalcF(bool initialize, bool calc_stresses) {
  if(calc_stresses) throw RUMD_Error("AxilrodTellerPotential",__func__,"Calculation of stresses not implemented");

  if(initialize)
    particleData->SetForcesToZero();

 
  neighborList.UpdateNBlist();

  if(testLESB)
    calcf_AxTe<<<kp.grid, kp.threads>>>( particleData->GetNumberOfParticles(), 
					 particleData->GetNumberOfVirtualParticles(),
					 particleData->d_r, particleData->d_f, 
					 particleData->d_w,
					 testLESB, testLESB->GetDevicePointer(),
					 v_AT,
					 neighborList.GetNumNbrsPtr(),
					 neighborList.GetNbListPtr());
  else
    calcf_AxTe<<<kp.grid, kp.threads>>>( particleData->GetNumberOfParticles(), 
					 particleData->GetNumberOfVirtualParticles(),
					 particleData->d_r, particleData->d_f, 
					 particleData->d_w,
					 testRSB, testRSB->GetDevicePointer(),
					 v_AT,
					 neighborList.GetNumNbrsPtr(),
					 neighborList.GetNbListPtr());

}




double AxilrodTellerPotential::GetPotentialEnergy(){

  neighborList.UpdateNBlist();
  unsigned nvp = particleData->GetNumberOfVirtualParticles();
  if(allocated_size_pe != nvp)
    AllocatePE_Array(nvp);

  // set local force, virial arrays to zero
  cudaMemset( d_f_loc, 0, nvp * sizeof(float4) );
  cudaMemset( d_w_loc, 0, nvp * sizeof(float4) ); // currently not used
  
  if(testLESB)
    calcf_AxTe<<<kp.grid, kp.threads>>>( particleData->GetNumberOfParticles(), 
					 nvp,
					 particleData->d_r, d_f_loc, 
					 d_w_loc,
					 testLESB, testLESB->GetDevicePointer(),
					 v_AT,
					 neighborList.GetNumNbrsPtr(),
					 neighborList.GetNbListPtr());
  else
    calcf_AxTe<<<kp.grid, kp.threads>>>( particleData->GetNumberOfParticles(), 
					 particleData->GetNumberOfVirtualParticles(),
					 particleData->d_r, d_f_loc, 
					 d_w_loc,
					 testRSB, testRSB->GetDevicePointer(),
					 v_AT,
					 neighborList.GetNumNbrsPtr(),
					 neighborList.GetNbListPtr());



  
  double pe_thisPotential = 0.;

  // copy to host
  cudaMemcpy( h_f_w_loc, d_f_loc, allocated_size_pe * sizeof(float4), cudaMemcpyDeviceToHost );

  // sum over particles
  for(unsigned int i=0; i < particleData->GetNumberOfParticles(); i++)
    pe_thisPotential += h_f_w_loc[i].w;
  
  return pe_thisPotential;
}




template<class S>
__global__ void calcf_AxTe(unsigned int num_part, unsigned int nvp, __restrict__ float4* r, __restrict__ float4* f, float4* w, S* simBox, float* simBoxPointer, float v_AT, unsigned *num_nbrs, unsigned *nbl) {
extern __shared__ float4 s_r[];
//float* s_params = (float*) &s_r[0]; // NOTE overlaps s_r. Should check size etc 

  float4 my_f = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_w = {0.0f, 0.0f, 0.0f, 0.0f};

  float simBoxPtr_local[simulationBoxSize];
  simBox->loadMemory( simBoxPtr_local, simBoxPointer );

 
  int my_num_nbrs = num_nbrs[MyGP];
  int nbj;
  int nbk;
  //int nbj_prefetch = nbl[MyGP];
  
  // 
  float4 r_i = (r[MyGP]);

  for (unsigned int j=0; j<my_num_nbrs; j++) {
    nbj = nbl[nvp*j + MyGP];
    float4 r_j = LOAD(r[nbj]);
    
    float4 dist_ij = simBox->calculateDistance(r_i, r_j, simBoxPtr_local);
    for (unsigned int k=j+1; k<my_num_nbrs; k++) {
      nbk = nbl[nvp*k + MyGP];
      float4 r_k = LOAD(r[nbk]);
      float4 dist_ik = simBox->calculateDistance(r_i, r_k, simBoxPtr_local);
      float4 dist_jk = simBox->calculateDistance(r_j, r_k, simBoxPtr_local);
      // float4 is a unit containing 4 floats
      // dist_ik.x = r_i.x - r_j.x (with PBC taken into account)
      // dist_ik.y = r_i.y - r_j.y (with PBC taken into account)
      // dist_ik.z = r_i.z - r_j.z (with PBC taken into account)
      // The fourth component, dist_ik.w is the sum of the squares of the others, ie the squared magnitude of the displacement vector (with PBC taken into account)
      
      
      // insert code to calculate the force on particle i  here


      // add force on particle i to global force array
      atomicAdd(&(f[MyGP].x), my_f.x);
      atomicAdd(&(f[MyGP].y), my_f.y);
      atomicAdd(&(f[MyGP].z), my_f.z);
      atomicAdd(&(f[MyGP].w), my_f.w);
      atomicAdd(&(w[MyGP].w), my_w.w);

    } // k loop

  } // j loop
  
}
